/*
 * file CqContextResource.java
 *
 * Licensed Materials - Property of IBM
 * Restricted Materials of IBM 
 *
 * com.ibm.rational.wvcm.stp.cq.CqContextResource
 *
 *  Copyright IBM Corporation 2004, 2008.  All Rights Reserved.
 * Note to U.S. Government Users Restricted Rights:  Use, duplication or 
 * disclosure restricted by GSA ADP  Schedule Contract with IBM Corp.
 */
package com.ibm.rational.wvcm.stp.cq;

import java.util.List;

import javax.wvcm.Feedback;
import javax.wvcm.Resource;
import javax.wvcm.WvcmException;
import javax.wvcm.PropertyNameList.PropertyName;

import com.ibm.rational.wvcm.stp.cq.CqProvider;
import com.ibm.rational.wvcm.stpex.StpExBase;
import com.ibm.rational.wvcm.stp.StpException;
import com.ibm.rational.wvcm.stp.StpResource;


/**
 * A CqContextResource is a ClearQuest user database resource that may be
 * temporarily held in a change context while it is being modified and before
 * those modifications are committed to the database. The server maintains a
 * separate and private change context for each database a client is logged
 * into. The new and modified resources in the change context are visible only
 * to the client and to the hooks the client fires and will not be visible to
 * other clients of the database until the resources in the change context are
 * delivered to the database.
 * 
 * <p>
 * The process of modifying context resources involves three steps:
 * </p>
 * 
 * <ul>
 * <li><b>Initiate:</b> A new context resource is created, one or more
 * properties of an existing context resource are updated, or an existing
 * context resource is deleted. For records and attachments, this initial step
 * requires the specification of an action to be used, thus declaring the
 * business rules to be followed in making the modifications.</li>
 * <li><b>Modify:</b> Modifications are made to the copy of the resource in
 * the change context. As changes are written to the server they are verified on
 * the server according to the business rules of the schema. Note that moribund
 * resources (resources in the change context that have been deleted) cannot be
 * modified.</li>
 * <li><b>Deliver:</b> When all resource modifications have been completed,
 * the modified resources are delivered to the user database to make changes
 * permanent.</li>
 * </ul>
 * 
 * <p>
 * This modification process allows the client to work with its user to make
 * coordinated changes to multiple resources, with the option of altering or
 * abandoning at any time changes to any of the resources involved. In its full
 * generality, this modification process could involve three or more
 * interactions with the server, but for simple cases, the API allows the steps
 * to be combined such that most modifications can be completed in only one
 * interaction with the server.
 * </p>
 * 
 * <p>
 * Once a modification has been initiated by a client for a user, changes made
 * to the resources involved are not visible to other users or clients until the
 * modifications are delivered to the user database. The changes (including
 * creation and deletion) are confined to the change context used and visible
 * only through proxies obtained from the provider that made the changes.
 * </p>
 * 
 * <p>
 * The precise locking semantics of this modification process depend on the type
 * of resource being modified. The only guarantee is that if an
 * initiate-modify-deliver sequence is successfully completed by a client, the
 * resources modified did not change in the database while they were being
 * modified by that client.
 * </p>
 * 
 * <p>
 * When the modification of a resource is initiated, a writable version of the
 * resource is created in the change context associated with the proxy used.
 * Unless the resource is being created, the properties of the original resource
 * are subsequently copied to this new version. Subsequent operations targeting
 * the original resource through a proxy from the same provider will be
 * redirected to operate on the version cached by the change context.
 * </p>
 * 
 * <p>
 * The delivery of modified resources from a change context to the database is
 * controlled by the delivery order list parameter of any "do" method designed
 * to operate on change context resources. Such "do" methods are specified in
 * this interface and its extensions. See also
 * {@link CqUserDb#doDeliver(Feedback, List)}.
 * <p>
 * The delivery order list parameter specifies which resources are to be
 * delivered after successful completion of the operation. The resources are
 * delivered in the order in which they appear in the delivery order list.
 * Unique delivery order list values are defined by {@link CqProvider}} for
 * specifying the delivery of the entire change context or just the resources
 * modified by the current operation.
 * </p>
 * <p>
 * Modifications (creations or deletions) initiated in a change context can be
 * abandoned before they are delivered by removing the modified resource from
 * the change context. "do" methods are defined for clearing the entire change
 * context ({@link CqUserDb#doClearContext(Feedback)}, reverting any
 * individual resource ({@link #doRevert(Feedback)} or reverting a list of
 * resources ({@link CqUserDb#doRevert(Feedback, List)}.
 * </p>
 * <p>
 * The {@link #doWriteProperties(Feedback, List)} and
 * {@link #doUnbindAll(Feedback, List)} methods in this interface overload
 * methods defined by the WVCM standard. If the standard methods are applied to
 * a context resource, they behave as if the {@link CqProvider#AUTO} delivery
 * order list was used in the extended method. If the targeted context resource
 * is not yet in the change context, {@link CqProvider#AUTO} behaves like the
 * {@link CqProvider#DELIVER}, causing any changes made by the method to be
 * immediately delivered from the change context. If the targeted resource is
 * already in the change context, {@link CqProvider#AUTO} behaves like
 * {@link CqProvider#HOLD}, leaving the modified resource still in the change
 * context. Since the generic WVCM client would not be aware of the change
 * context, it looks to the WVCM client as if doWriteProperties is simply
 * writing properties directly to the resource on the server.
 * </p>
 * 
 * @see CqUserDb#doClearContext(Feedback)
 * @see CqUserDb#doDeliver(Feedback, List)
 * @see CqUserDb#doRevert(Feedback, List)
 * @see CqUserDb#getModifiedResourcesList()
 * @see CqUserDb#getMoribundResourcesList()
 * @see CqProvider#HOLD
 * @see CqProvider#DELIVER
 * @see CqProvider#DELIVER_ALL
 * @see CqProvider#AUTO
 * 
 */
public interface CqContextResource extends CqUserDbMember
{
    /**
     * Attempts to deliver this resource to the database.
     * 
     * @param feedback A request for the properties of this resource that are to
     *            be included in the proxy returned by this operation.
     * @return A proxy for the delivered resource populated with the requested
     *         properties.
     * @throws WvcmException if this resource has no modifications to deliver to
     *             the database or if any of the modifications are invalid.
     */
    CqContextResource doDeliver(Feedback feedback) throws WvcmException;
    
    /**
     * Removes the modified or moribund resource referenced by this proxy from
     * the client's change context, thereby revealing at that location the
     * resource, if any, in the database. This is <i>not</i> an undo operation
     * applied only to the most recent modification of the resource. If removes
     * <i>all</i> modifications made to the resource since it was last
     * delivered to the database.
     * <p>
     * Note: other resources in the change context, even though initiated by
     * changes to this resource, are not removed from the change context by this
     * operation.
     * <P>
     * Since the target of this operation is being removed it makes little sense
     * to write dirty properties to that resource before it is removed.
     * Consequently, any dirty properties in the proxy are ignored.
     * 
     * @param feedback A Feedback object requesting property values from the
     *            resource after the resource has been reverted. May be <b>null</b>
     *            if no properties are desired.
     * @return A proxy for the resource made visible after removing the modified
     *         or moribund resource from the change context. Will be <b>null</b>
     *         if no such resource exists.
     * 
     * @throws WvcmException if the resource was not writable or if there were
     *             problems reverting it.
     */
    CqContextResource doRevert(Feedback feedback)
        throws WvcmException;

    /**
     * Writes dirty properties from this proxy to the targeted context resource
     * using implicitly {@link CqProvider#AUTO} as the delivery order list value.
     * <p>
     * This method is intended primarily for the use of generic WVCM clients,
     * who have no knowledge of change contexts and their semantics and would
     * expect context resources to behave simply as non-versioned resources. The
     * net impact of this operation on the change context is nil; it will
     * neither add new resources to the change context nor remove existing
     * resources from the change context.
     * <p>
     * If the resource to be modified is not in the change context prior to
     * calling this method, then the modified resource is immediately written
     * back to the database when the property update succeeds and is removed
     * from the change context when the update fails.
     * <p>
     * If the resource is already modified in the change context, then the
     * resource remains in the change context after the operation whether the
     * property update succeeds or fails.
     * <p>
     * If this proxy contains no dirty properties, no action is started, and no
     * delivery takes place. If, in addition, the PropertyRequest for the result
     * proxy provided by the feedback argument is null, no interaction with the
     * server is initiated and, hence, the proxy location is not verified. A
     * non-null (but possibly empty) PropertyRequest for the result proxy will
     * force a server interaction and verification of the proxy location.
     * 
     * @return A proxy for this resource (after delivery if a delivery happens)
     *         populated with the properties requested by the Feedback object
     *         for the result.
     * 
     * @throws WvcmException If there are errors when attempting update
     *             properties or deliver the resource back to the database.
     */
    Resource doWriteProperties(Feedback feedback)
        throws WvcmException;

    /**
     * Writes dirty properties to the change context-copy of this resource and,
     * optionally, requests that modified or moribund resources in the change
     * context be delivered to or deleted from the database.
     * <p>
     * If the change context does not already have a copy of this resource, one
     * is created in the change context. If an action is required and one is not
     * specified by a {@link CqRecord#ACTION} property value in this proxy, a
     * MODIFY-type action is started if one is uniquely defined for the
     * resource.
     * <p>
     * Updated properties are not written back to the database until delivery of
     * those changes to the database is requested by the client through the use
     * of a delivery order list parameter or invocation of a doDeliver method.
     * <p>
     * If this proxy contains no dirty properties, no action is started and the
     * state of the resource in the change context is not modified. If, in
     * addition, the PropertyRequest for the result proxy provided by the
     * feedback argument is <b>null</b> and also not delivery is requested, no
     * interaction with the server is initiated and, hence, the proxy location
     * is not verified.
     * <p>
     * A non-null (but possibly empty) PropertyRequest for the result proxy will
     * force a server interaction and verification of the proxy location, even
     * if the proxy contains no dirty properties and no delivery is indicated.
     * 
     * @param feedback An instance of Feedback requesting properties for this
     *            resource and for the resources delivered by this operation.
     *            May be <b>null</b> if no feedback is desired.
     * 
     * @param deliveryOrder If {@link CqProvider#HOLD} the modified resource is
     *            left in a writable state in the change context--the modified
     *            resource in the change context must be delivered before the
     *            modifications become visible to other providers. If not
     *            {@link CqProvider#HOLD}, the modified and moribund resources
     *            specified by this parameter will be delivered to or deleted
     *            from the database in the order indicated. To deliver all
     *            modified and moribund resources in an arbitrary order, use
     *            {@link CqProvider#DELIVER_ALL}. To deliver just this modified
     *            resource, use {@link CqProvider#DELIVER}. Must not be <b>null</b>.
     * @return A proxy for this resource populated with the properties requested
     *         by the Feedback object for the result. The properties reflect the
     *         state of the resource after delivery if delivery has been
     *         requested.
     * 
     * @throws WvcmException If there are errors when attempting to write
     *             properties or deliver the resource.
     */
    CqContextResource doWriteProperties(Feedback feedback,
                                        List<CqContextResource> deliveryOrder)
        throws WvcmException;

    /**
     * Initiates the deletion of this context resource from the database using
     * implicitly {@link CqProvider#AUTO} as the delivery order list value.
     * The deletion of the resource becomes permanent and visible to other
     * proxies only after this deletion has been delivered. Prior to delivery,
     * the deletion of the resource may be canceled by executing
     * {@link #doRevert(Feedback)} on the resource.
     * <p>
     * This method is intended primarily for the use of generic WVCM clients,
     * who have no knowledge of change contexts and their semantics and would
     * expect context resources to behave simply as non-versioned resources. The
     * net impact of this operation on the change context is nil; it will
     * neither add new resources to the change context nor remove existing
     * resources from the change context.
     * <p>
     * Between the time a resource is deleted and that deletion becomes
     * permanent, the resource is said to be <i>moribund</i> (near death). A
     * moribund resource can be targeted only by a {@link #doDeliver(Feedback)}
     * or {@link #doRevert(Feedback)} operation. Any other method targeted to a
     * moribund resource will throw a RESOURCE_NOT_FOUND exception. Moribund
     * resources may also appear in a delivery order list parameter.
     * <p>
     * If an action must be started to perform the deletion, another action must
     * not already be active on this resource. If no action is specified by a
     * {@link CqRecord#ACTION} property value in this proxy and a DELETE-type
     * action is not uniquely defined for this type of resource an
     * {@link com.ibm.rational.wvcm.stp.StpException.StpReasonCode#UNKNOWN_ACTION} exception is thrown.
     * <p>
     * If the resource was not in the change context when this method was
     * invoked, the deletion becomes permanent immediately--before the method
     * returns. If it was already in the change context, however, then no
     * delivery is attempted and delivery will have to be requested directly to
     * complete the deletion.
     * 
     * @throws WvcmException If the resource does not exist or cannot be deleted
     *             using the default DELETE action.
     * 
     * @see StpResource#doUnbindAll(Feedback)
     */
    void doUnbindAll(Feedback feedback)
        throws WvcmException;

    /**
     * Initiates the deletion of this context resource from the database. The
     * deletion of the resource becomes permanent and visible to other proxies
     * only after this deletion has been delivered. Prior to delivery, the
     * deletion of the resource may be canceled by executing
     * {@link #doRevert(Feedback)} on the resource.
     * <p>
     * Between the time a resource is deleted and that deletion becomes
     * permanent, the resource is said to be <i>moribund</i> (near death). A
     * moribund resource can be accessed only by a {@link #doDeliver(Feedback)}
     * or {@link #doRevert(Feedback)} operation. Any other method will throw a
     * RESOURCE_NOT_FOUND exception.
     * <p>
     * If an action must be started to perform the deletion, another action must
     * not already be active on this resource. If no action is specified by a
     * {@link CqRecord#ACTION} property value in this proxy and a DELETE-type
     * action is not uniquely defined for this type of resource an
     * {@link com.ibm.rational.wvcm.stp.StpException.StpReasonCode#UNKNOWN_ACTION} exception is thrown.
     * <p>
     * The disposition of the change context after the deletion succeeds is
     * controlled by the deliveryOrder parameter
     * </p>
     * 
     * @param feedback A Feedback object requesting the properties desired from
     *            resources modified during this operation. May be null if no
     *            properties are desired.
     * @param deliveryOrder If {@link CqProvider#HOLD} the moribund resource is
     *            left in the change context and not actually deleted--it must
     *            be delivered before the deletion becomes visible to other
     *            providers. If not {@link CqProvider#HOLD}, the modified and
     *            moribund resources specified by this parameter will be
     *            delivered to or deleted from the database in the order
     *            indicated. To deliver all modified and moribund resources in
     *            an arbitrary order, use {@link CqProvider#DELIVER_ALL}. To
     *            deliver just this moribund resource, use
     *            {@link CqProvider#DELIVER}. Must not be <b>null</b>.
     * 
     * @throws WvcmException If the resource does not exist or cannot be deleted
     *             using the default DELETE action.
     * 
     * @see StpResource#doUnbindAll(Feedback)
     */
    void doUnbindAll(Feedback feedback, 
                     List<CqContextResource> deliveryOrder)
        throws WvcmException;
   
    /**
     * True if a modified copy of this resource exists in this clients database
     * change context.
     */
    PropertyName<Boolean> IS_MODIFIED =
        new PropertyName<Boolean>(StpExBase.PROPERTY_NAMESPACE, "is-modified" /*NOI18N*/);

    /**
     * Returns the value of the {@link #IS_MODIFIED IS_MODIFIED} property as
     * defined by this proxy.
     * 
     * @return <b>true</b> if this proxy refers to a resource whose properties
     *         were modified in this client's change context; otherwise <b>false</b>.
     * 
     * @throws WvcmException if this proxy does not define a value for the
     *             {@link #IS_MODIFIED IS_MODIFIED} property.
     */
    boolean getIsModified()
        throws WvcmException;
}
